@@ -9,13 +9,12 @@ import sys
 from os.path import join, dirname, abspath, exists, splitext, basename
 import re
 from glob import glob
-from pprint import pprint
 import unittest
 import codecs
 import difflib
 import doctest
-
-from testlib import TestError, TestSkipped, tag
+from json import loads as json_loads
+import warnings
 
 sys.path.insert(0, join(dirname(dirname(abspath(__file__)))))
 try:
@@ -23,9 +22,13 @@ try:
 finally:
     del sys.path[0]
 
+#---- Test cases
 
 class _MarkdownTestCase(unittest.TestCase):
     """Helper class for Markdown tests."""
+
+    maxDiff = None
+
     def _assertMarkdownParity(self, text):
         """Assert that markdown2.py produces same output as Markdown.pl."""
         #TODO add normalization
@@ -38,60 +41,41 @@ class _MarkdownTestCase(unittest.TestCase):
                 == perl_html.replace('\n', '')):
             close_though = " (close though -- all but EOLs match)"
 
-        self.assertEqual(python_html, perl_html, _dedent(u"""\
+        self.assertEqual(python_html, perl_html, _dedent("""\
             markdown2.py didn't produce the same output as Markdown.pl%s:
               ---- text ----
             %s  ---- Python markdown2.py HTML ----
             %s  ---- Perl Markdown.pl HTML ----
-            %s""") % (close_though, _display(text), 
+            %s""") % (close_though, _display(text),
                       _display(python_html), _display(perl_html)))
 
-    _xml_escape_re = re.compile(r'&#(x[0-9A-Fa-f]{2,3}|[0-9]{2,3});')
-    def _xml_escape_sub(self, match):
-        escape = match.group(1)
-        if escape[0] == 'x':
-            return unichr(int('0'+escape, base=16))
-        else:
-            return unichr(int(escape))
-
-    _markdown_email_link_re = re.compile(ur'<a href="(.*?&#.*?)">(.*?)</a>', re.U)
-    def _markdown_email_link_sub(self, match):
-        href, text = match.groups()
-        href = self._xml_escape_re.sub(self._xml_escape_sub, href)
-        text = self._xml_escape_re.sub(self._xml_escape_sub, text)
-        return '<a href="%s">%s</a>' % (href, text)
-
-    def _normalize(self, html):
-        """Normalize (somewhat) Markdown'd HTML.
-
-        Part of Markdown'ing involves obfuscating email links with
-        randomize encoding. Undo that obfuscation.
-        
-        Also normalize EOLs.
-        """
-        html = self._markdown_email_link_re.sub(
-            self._markdown_email_link_sub, html)
-        if sys.platform == "win32":
-            html = html.replace('\r\n', '\n')
-        return html
-
-    def _assertMarkdownPath(self, text_path, encoding="utf-8", opts=None):
+    def _assertMarkdownPath(self, text_path, encoding="utf-8", opts=None,
+            toc_html_path=None, metadata_path=None):
         text = codecs.open(text_path, 'r', encoding=encoding).read()
         html_path = splitext(text_path)[0] + ".html"
         html = codecs.open(html_path, 'r', encoding=encoding).read()
-        self._assertMarkdown(text, html, text_path, html_path, opts=opts)
+        extra = {}
+        if toc_html_path:
+            extra["toc_html"] = codecs.open(toc_html_path, 'r', encoding=encoding).read()
+            extra["toc_html_path"] = toc_html_path
+        if metadata_path:
+            extra["metadata"] = json_loads(
+                codecs.open(metadata_path, 'r', encoding=encoding).read())
+            extra["metadata_path"] = metadata_path
+        self._assertMarkdown(text, html, text_path, html_path, opts=opts, **extra)
 
     def _assertMarkdown(self, text, html, text_path=None, html_path=None,
-                        opts=None):
+            opts=None, toc_html=None, toc_html_path=None, metadata=None,
+            metadata_path=None):
         """Assert that markdown2.py produces the expected HTML."""
         if text_path is None: text_path = "<text content>"
         if html_path is None: html_path = "<html content>"
         if opts is None:
             opts = {}
 
-        norm_html = self._normalize(html)
+        norm_html = norm_html_from_html(html)
         python_html = markdown2.markdown(text, **opts)
-        python_norm_html = self._normalize(python_html)
+        python_norm_html = norm_html_from_html(python_html)
 
         close_though = ""
         if python_norm_html != norm_html \
@@ -99,34 +83,60 @@ class _MarkdownTestCase(unittest.TestCase):
                 == norm_html.replace('\n', '')):
             close_though = " (close though -- all but EOLs match)"
 
-        diff = u''
+        diff = ''
         if python_norm_html != norm_html:
             diff = difflib.unified_diff(
-                    norm_html.splitlines(1), 
+                    norm_html.splitlines(1),
                     python_norm_html.splitlines(1),
                     html_path,
                     "markdown2 "+text_path)
-            diff = u''.join(diff)
-        errmsg = _dedent(u"""\
+            diff = ''.join(diff)
+        errmsg = _dedent("""\
             markdown2.py didn't produce the expected HTML%s:
               ---- text (escaping: .=space, \\n=newline) ----
             %s  ---- Python markdown2.py HTML (escaping: .=space, \\n=newline) ----
             %s  ---- expected HTML (escaping: .=space, \\n=newline) ----
             %s  ---- diff ----
-            %s""") % (close_though, _display(text), 
+            %s""") % (close_though, _display(text),
                       _display(python_html), _display(html),
                       _indent(diff))
 
         def charreprreplace(exc):
             if not isinstance(exc, UnicodeEncodeError):
                 raise TypeError("don't know how to handle %r" % exc)
-            # repr -> remote "u'" and "'"
-            obj_repr = repr(exc.object[exc.start:exc.end])[2:-1]
-            return (unicode(obj_repr), exc.end)
+            obj_repr = repr(exc.object[exc.start:exc.end])[1:-1]
+            return (str(obj_repr), exc.end)
         codecs.register_error("charreprreplace", charreprreplace)
 
-        self.assertEqual(python_norm_html, norm_html, 
-                         errmsg.encode('ascii', 'charreprreplace'))
+        self.assertEqual(python_norm_html, norm_html, errmsg)
+
+        if toc_html:
+            python_toc_html = python_html.toc_html
+            python_norm_toc_html = norm_html_from_html(python_toc_html)
+            norm_toc_html = norm_html_from_html(toc_html)
+
+            diff = ''
+            if python_norm_toc_html != norm_toc_html:
+                diff = difflib.unified_diff(
+                        norm_toc_html.splitlines(1),
+                        python_norm_toc_html.splitlines(1),
+                        toc_html_path,
+                        "`markdown2 %s`.toc_html" % text_path)
+                diff = ''.join(diff)
+            errmsg = _dedent("""\
+                markdown2.py didn't produce the expected TOC HTML%s:
+                  ---- text (escaping: .=space, \\n=newline) ----
+                %s  ---- Python markdown2.py TOC HTML (escaping: .=space, \\n=newline) ----
+                %s  ---- expected TOC HTML (escaping: .=space, \\n=newline) ----
+                %s  ---- diff ----
+                %s""") % (close_though, _display(text),
+                          _display(python_toc_html), _display(toc_html),
+                          _indent(diff))
+            self.assertEqual(python_norm_toc_html, norm_toc_html,
+                errmsg.encode('utf-8', 'charreprreplace'))
+
+        if metadata:
+            self.assertEqual(python_html.metadata, metadata)
 
     def generate_tests(cls):
         """Add test methods to this class for each test file in
@@ -141,13 +151,26 @@ class _MarkdownTestCase(unittest.TestCase):
             opts_path = splitext(text_path)[0] + ".opts"
             if exists(opts_path):
                 try:
-                    opts = eval(open(opts_path, 'r').read())
-                except Exception, ex:
-                    print "WARNING: couldn't load `%s' opts file: %s" \
-                          % (opts_path, ex)
+                    with warnings.catch_warnings(record=True) as caught_warnings:
+                        opts = eval(open(opts_path, 'r').read())
+                    for warning in caught_warnings:
+                        print("WARNING: loading %s generated warning: %s - lineno %d" % (opts_path, warning.message, warning.lineno), file=sys.stderr)
+                except Exception:
+                    _, ex, _ = sys.exc_info()
+                    print("WARNING: couldn't load `%s' opts file: %s" \
+                          % (opts_path, ex), file=sys.stderr)
 
-            test_func = lambda self, t=text_path, o=opts: \
-                self._assertMarkdownPath(t, opts=o)
+            toc_html_path = splitext(text_path)[0] + ".toc_html"
+            if not exists(toc_html_path):
+                toc_html_path = None
+            metadata_path = splitext(text_path)[0] + ".metadata"
+            if not exists(metadata_path):
+                metadata_path = None
+
+            test_func = lambda self, t=text_path, o=opts, c=toc_html_path, \
+                    m=metadata_path: \
+                    self._assertMarkdownPath(t, opts=o, toc_html_path=c,
+                          metadata_path=m)
 
             tags_path = splitext(text_path)[0] + ".tags"
             if exists(tags_path):
@@ -179,12 +202,11 @@ class PHPMarkdownTestCase(_MarkdownTestCase):
 
 class PHPMarkdownExtraTestCase(_MarkdownTestCase):
     """Test cases from MDTest.
-    
+
     These are all knownfailures because these test non-standard Markdown
     syntax no implemented in markdown2.py.  See
     <http://www.michelf.com/projects/php-markdown/extra/> for details.
     """
-    __tags__ = ["knownfailure"]
     cases_dir = "php-markdown-extra-cases"
 
 
@@ -193,6 +215,32 @@ class DirectTestCase(_MarkdownTestCase):
     Python-markdown (markdown.py).
     """
 
+    def test_slow_hr(self):
+        import time
+        text = """\
+* * *
+
+This on *almost* looks like an hr, except for the trailing '+'. In older
+versions of markdown2.py this was pathologically slow:
+
+- - - - - - - - - - - - - - - - - - - - - - - - - +
+"""
+        html = """\
+<hr />
+
+<p>This on <em>almost</em> looks like an hr, except for the trailing '+'. In older
+versions of markdown2.py this was pathologically slow:</p>
+
+<p>- - - - - - - - - - - - - - - - - - - - - - - - - +</p>
+"""
+        start = time.time()
+        self._assertMarkdown(text, html)
+        end = time.time()
+        delta = end - start
+        self.assertTrue(delta < 1.0, "It took more than 1s to process "
+            "'slow-hr'. It took %.2fs. Too slow!" % delta)
+    test_slow_hr.tags = ["perf"]
+
     def test_code_in_strong(self):
         self._assertMarkdown(
             '**look at `this code` call**',
@@ -208,39 +256,110 @@ class DirectTestCase(_MarkdownTestCase):
     def test_pre(self):
         self._assertMarkdown(_dedent('''\
             some starter text
-            
+
                 #!/usr/bin/python
                 print "hi"'''),
             '<p>some starter text</p>\n\n<pre><code>#!/usr/bin/python\nprint "hi"\n</code></pre>\n')
 
     def test_russian(self):
-        ko = u'\u043b\u0449' # 'ko' on russian keyboard
-        self._assertMarkdown(u"## %s" % ko,
-            u'<h2>%s</h2>\n' % ko)
+        ko = '\\u043b\\u0449' # 'ko' on russian keyboard
+        self._assertMarkdown("## %s" % ko,
+            '<h2>%s</h2>\n' % ko)
     test_russian.tags = ["unicode", "issue3"]
 
+    def test_toc_with_persistent_object(self):
+        """
+        Tests that the toc is the same every time it's run on HTML, even if the Markdown object isn't disposed of.
+        """
+        md = markdown2.Markdown(extras=["toc"])
+        html = """
+# Header 1
+## Header 1.1
+## Header 1.2
+### Header 1.3
+# Header 2
+## Header 2.1
+        """
+        expected_toc_html = """<ul>
+  <li><a href="#header-1">Header 1</a>
+  <ul>
+    <li><a href="#header-11">Header 1.1</a></li>
+    <li><a href="#header-12">Header 1.2</a>
+    <ul>
+      <li><a href="#header-13">Header 1.3</a></li>
+    </ul></li>
+  </ul></li>
+  <li><a href="#header-2">Header 2</a>
+  <ul>
+    <li><a href="#header-21">Header 2.1</a></li>
+  </ul></li>
+</ul>
+"""
+        self.assertEqual(expected_toc_html, md.convert(html).toc_html)
+        # Do it again, to check if the toc_html is just appended rather than replaced
+        self.assertEqual(expected_toc_html, md.convert(html).toc_html)
+        # Create different html, and confirm toc_html is replaced
+        html = """
+# I'm new html
+## I don't have to be long, just different
+"""
+        expected_toc_html = """<ul>
+  <li><a href="#im-new-html">I'm new html</a>
+  <ul>
+    <li><a href="#i-dont-have-to-be-long-just-different">I don't have to be long, just different</a></li>
+  </ul></li>
+</ul>
+"""
+        self.assertEqual(expected_toc_html, md.convert(html).toc_html)
+    test_toc_with_persistent_object.tags = ["toc", "issue208"]
+
 
 class DocTestsTestCase(unittest.TestCase):
     def test_api(self):
-        if sys.version_info[:2] < (2,4):
-            raise TestSkipped("no DocFileTest in Python <=2.3")
         test = doctest.DocFileTest("api.doctests")
         test.runTest()
 
-    def test_internal(self):
-        doctest.testmod(markdown2)
-
-
 
 #---- internal support stuff
 
+_xml_escape_re = re.compile(r'&#(x[0-9A-Fa-f]{2,3}|[0-9]{2,3});')
+def _xml_escape_sub(match):
+    escape = match.group(1)
+    if escape[0] == 'x':
+        return chr(int('0'+escape, base=16))
+    else:
+        return chr(int(escape))
+
+_markdown_email_link_re = re.compile(r'<a href="(.*?&#.*?)">(.*?)</a>', re.U)
+def _markdown_email_link_sub(match):
+    href, text = match.groups()
+    href = _xml_escape_re.sub(_xml_escape_sub, href)
+    text = _xml_escape_re.sub(_xml_escape_sub, text)
+    return '<a href="%s">%s</a>' % (href, text)
+
+def norm_html_from_html(html):
+    """Normalize (somewhat) Markdown'd HTML.
+
+    Part of Markdown'ing involves obfuscating email links with
+    randomize encoding. Undo that obfuscation.
+
+    Also normalize EOLs.
+    """
+    if not isinstance(html, str):
+        html = html.decode('utf-8')
+    html = _markdown_email_link_re.sub(
+        _markdown_email_link_sub, html)
+    html = html.replace('\r\n', '\n')
+    return html
+
+
 def _display(s):
     """Markup the given string for useful display."""
-    if not isinstance(s, unicode):
+    if not isinstance(s, str):
         s = s.decode("utf-8")
     s = _indent(_escaped_text_from_text(s, "whitespace"), 4)
     if not s.endswith('\n'):
-        s += u'\n'
+        s += '\n'
     return s
 
 def _markdown_with_perl(text):
@@ -261,20 +380,20 @@ def _markdown_with_perl(text):
 # Recipe: dedent (0.1.2)
 def _dedentlines(lines, tabsize=8, skip_first_line=False):
     """_dedentlines(lines, tabsize=8, skip_first_line=False) -> dedented lines
-    
+
         "lines" is a list of lines to dedent.
         "tabsize" is the tab width to use for indent width calculations.
         "skip_first_line" is a boolean indicating if the first line should
             be skipped for calculating the indent width and for dedenting.
             This is sometimes useful for docstrings and similar.
-    
+
     Same as dedent() except operates on a sequence of lines. Note: the
     lines list is modified **in-place**.
     """
     DEBUG = False
-    if DEBUG: 
-        print "dedent: dedent(..., tabsize=%d, skip_first_line=%r)"\
-              % (tabsize, skip_first_line)
+    if DEBUG:
+        print("dedent: dedent(..., tabsize=%d, skip_first_line=%r)"\
+              % (tabsize, skip_first_line))
     indents = []
     margin = None
     for i, line in enumerate(lines):
@@ -291,12 +410,12 @@ def _dedentlines(lines, tabsize=8, skip_first_line=False):
                 break
         else:
             continue # skip all-whitespace lines
-        if DEBUG: print "dedent: indent=%d: %r" % (indent, line)
+        if DEBUG: print("dedent: indent=%d: %r" % (indent, line))
         if margin is None:
             margin = indent
         else:
             margin = min(margin, indent)
-    if DEBUG: print "dedent: margin=%r" % margin
+    if DEBUG: print("dedent: margin=%r" % margin)
 
     if margin is not None and margin > 0:
         for i, line in enumerate(lines):
@@ -308,7 +427,7 @@ def _dedentlines(lines, tabsize=8, skip_first_line=False):
                 elif ch == '\t':
                     removed += tabsize - (removed % tabsize)
                 elif ch in '\r\n':
-                    if DEBUG: print "dedent: %r: EOL -> strip up to EOL" % line
+                    if DEBUG: print("dedent: %r: EOL -> strip up to EOL" % line)
                     lines[i] = lines[i][j:]
                     break
                 else:
@@ -316,8 +435,8 @@ def _dedentlines(lines, tabsize=8, skip_first_line=False):
                                      "line %r while removing %d-space margin"
                                      % (ch, line, margin))
                 if DEBUG:
-                    print "dedent: %r: %r -> removed %d/%d"\
-                          % (line, ch, removed, margin)
+                    print("dedent: %r: %r -> removed %d/%d"\
+                          % (line, ch, removed, margin))
                 if removed == margin:
                     lines[i] = lines[i][j+1:]
                     break
@@ -337,7 +456,7 @@ def _dedent(text, tabsize=8, skip_first_line=False):
         "skip_first_line" is a boolean indicating if the first line should
             be skipped for calculating the indent width and for dedenting.
             This is sometimes useful for docstrings and similar.
-    
+
     textwrap.dedent(s), but don't expand tabs to spaces
     """
     lines = text.splitlines(1)
@@ -382,8 +501,8 @@ def _escaped_text_from_text(text, escapes="eol"):
     # - Add 'c-string' style.
     # - Add _escaped_html_from_text() with a similar call sig.
     import re
-    
-    if isinstance(escapes, basestring):
+
+    if isinstance(escapes, str):
         if escapes == "eol":
             escapes = {'\r\n': "\\r\\n\r\n", '\n': "\\n\n", '\r': "\\r\r"}
         elif escapes == "whitespace":
@@ -398,13 +517,8 @@ def _escaped_text_from_text(text, escapes="eol"):
 
     # Sort longer replacements first to allow, e.g. '\r\n' to beat '\r' and
     # '\n'.
-    escapes_keys = escapes.keys()
-    try:
-        escapes_keys.sort(key=lambda a: len(a), reverse=True)
-    except TypeError:
-        # Python 2.3 support: sort() takes no keyword arguments
-        escapes_keys.sort(lambda a,b: cmp(len(a), len(b)))
-        escapes_keys.reverse()
+    escapes_keys = list(escapes.keys())
+    escapes_keys.sort(key=lambda a: len(a), reverse=True)
     def repl(match):
         val = escapes[match.group(0)]
         return val
@@ -417,7 +531,7 @@ def _escaped_text_from_text(text, escapes="eol"):
 def _one_line_summary_from_text(text, length=78,
         escapes={'\n':"\\n", '\r':"\\r", '\t':"\\t"}):
     r"""Summarize the given text with one line of the given length.
-    
+
         "text" is the text to summarize
         "length" (default 78) is the max length for the summary
         "escapes" is a mapping of chars in the source text to
@@ -450,4 +564,3 @@ def test_cases():
     yield PHPMarkdownExtraTestCase
     yield DirectTestCase
     yield DocTestsTestCase
-
